#ifndef __MAP_H__
#define __MAP_H__
#include <vector>
#include <string>
#include <utility>
#include "constants.h"
#include "display.h"
#include "position.h"
#include "rng.h"

class game_map final {
private:
        static const int MAX_ROOM_CNT      = 20;
        static const int MIN_ROOM_WIDTH    = 3;
        static const int MIN_ROOM_HEIGHT   = 3;
        static const int MAX_ROOM_WIDTH    = 20;
        static const int MAX_ROOM_HEIGHT   = 5;
        // setup safezones
        static const int ACTUAL_MAP_WIDTH  = MAP_WIDTH - 4;
        static const int ACTUAL_MAP_HEIGHT = MAP_HEIGHT - 4;
        static const int MAP_PADDING       = 2;
        static const int MIN_ROOM_SPACING  = 3;
        static const int WIDTH_RESERVED    = 8;
        static const int HEIGHT_RESERVED   = 4;

        const feature enabled_features;
        std::vector<char> map;
        // IMPORTANT: keep empty_spots ordered
        position_list empty_spots;
        position up_stairs;
        position down_stairs;
        int room_cnt;
public:
        // randomly generate a map
        game_map(RNG *rng, const feature enabled_features);
        // read map from a string
        game_map(const std::string &map_data, RNG *rng,
                 const feature enabled_features);

        const position_list get_available_positions() const;
        bool is_available(const position &pos) const;

        const std::vector<position_list> get_room_list() const;
        // IMPORTANT: always print a map before anything else
        void print(display &out) const;

        position get_up_stairs() const;
        position get_down_stairs() const;

        int get_up_stairs_room() const;
        int get_down_stairs_room() const;
private:
        position remap_index(const int idx) const {
                return {idx % MAP_WIDTH, idx / MAP_HEIGHT};
        }

        int remap_position(const position &pos) const {
                return pos.y * MAP_WIDTH + pos.x;
        }

        position random_size(int min_width, int min_height,
                             int max_width, int max_height,
                             RNG *rng);

        std::vector<std::pair<position, int>> gen_room_dims(RNG *rng);
        std::vector<std::pair<position, position>> distr_rooms(RNG *rng,
                                                std::vector<std::pair<position, int>>
                                                &room_dims);

        bool overlap_x(std::pair<position, position> &room1,
                       std::pair<position, position> &room2);
        bool overlap_y(std::pair<position, position> &room1,
                       std::pair<position, position> &room2);
        void jitter(RNG *rng, std::vector<std::pair<position, position>> &rooms);
};

const std::string default_map =
        "|-----------------------------------------------------------------------------|\
|                                                                             |\
| |--------------------------|        |-----------------------|               |\
| |11111111111111111111111111|        |33333333333333333333333|               |\
| |11111111111111111111111111+########+33333333333333333333333|-------|       |\
| |11111111111111111111111111|   #    |3333333333333333333333333333333|--|    |\
| |11111111111111111111111111|   #    |3333333333333333333333333333333333|--| |\
| |----------+---------------|   #    |----+----------------|333333333333333| |\
|            #                 #############                |333333333333333| |\
|            #                 #     |-----+------|         |333333333333333| |\
|            #                 #     |444444444444|         |333333333333333| |\
|            ###################     |444444444444|   ######+333333333333333| |\
|            #                 #     |444444444444|   #     |333333333333333| |\
|            #                 #     |-----+------|   #     |--------+------| |\
|  |---------+-----------|     #           #          #              #        |\
|  |222222222222222222222|     #           #          #         |----+------| |\
|  |222222222222222222222|     ########################         |00000000000| |\
|  |222222222222222222222|     #           #                    |00000000000| |\
|  |222222222222222222222|     #    |------+--------------------|00000000000| |\
|  |222222222222222222222|     #    |000000000000000000000000000000000000000| |\
|  |222222222222222222222+##########+000000000000000000000000000000000000000| |\
|  |222222222222222222222|          |000000000000000000000000000000000000000| |\
|  |---------------------|          |---------------------------------------| |\
|                                                                             |\
|-----------------------------------------------------------------------------|";

#endif
